!+
! Module: edusoft_mod
!+
! Module: edusoft_mod
!
! Authors:  J. Ballet - SAp  CEA Saclay
!          
! Description: Declarations for the edusoft suite
! 
! Comments: 
! 
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module edusoft_mod

  use types

  implicit none 

  private

  public :: es_nmax, edu_npat, edu_nsid
  public :: edu_pattern, edu_evt, edu_out, sas_evt
  public :: projectEventsCounts, projectEventsEnergy

    integer, parameter :: es_nmax  = 50000 ! = EDU_NEVTMAX in edu_simu.h
    integer, parameter :: edu_npat = 32    ! number of EDU patterns
    integer, parameter :: edu_nsid = 5     ! size of the EDU window
    type edu_pattern
       integer               id_patt
       integer               id_mask
       integer(kind=int8)    mat(edu_nsid,edu_nsid)
       integer(kind=int8)    number
       integer(kind=int8)    id_kind
    end type edu_pattern
    type edu_evt
       integer(kind=int16)   x
       integer(kind=int16)   y
       integer(kind=int16)   pattern
       integer(kind=int16)   e1
       integer(kind=int16)   e2
       integer(kind=int16)   e3
       integer(kind=int16)   e4
       integer(kind=int16)   peripix
    end type edu_evt
    type edu_out
       integer         nevent
       type(edu_evt)   evt(es_nmax)
       integer         npix
       integer         fifoovf
    end type edu_out
    type sas_evt
       integer(kind=int8)  pattern
       integer(kind=int8)  peripix
       integer(kind=int16) rawx
       integer(kind=int16) rawy
       integer(kind=int32) frame
       integer(kind=int32) flag
       integer(kind=int16) e1
       integer(kind=int16) e2
       integer(kind=int16) e3
       integer(kind=int16) e4
    end type sas_evt

contains

! Project the pixels above threshold of an array of events onto an image

  subroutine projectEventsCounts(evt, patabove, image)

! evt        : Array of event structure
! patabove   : Geometry of event outside central pixel
! image      : Image upon which to project

    type(sas_evt), dimension(:), intent(in) :: evt
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1),  &
                   intent(in)    :: patabove
    integer(kind=int32), dimension(-2:,-2:), intent(inout) :: image

    integer :: x0, y0, kr, nevents

! Loop over events

    nevents = size(evt)
    do kr = 1,nevents
       x0 = evt(kr)%rawx - 2          ! Center of the event
       y0 = evt(kr)%rawy - 2
       image(x0,y0) = image(x0,y0) + 1
       image(x0-1:x0+1,y0-1:y0+1) = image(x0-1:x0+1,y0-1:y0+1) &
                                  + patabove(:,:,evt(kr)%pattern)
    enddo

  end subroutine projectEventsCounts

! Project the energy of an array of events onto an image

  subroutine projectEventsEnergy(evt, patabove, image)

! evt        : Array of event structure
! patabove   : Geometry of event outside central pixel
! image      : Image upon which to project

    type(sas_evt), dimension(:), intent(in) :: evt
    integer(kind=int16), dimension(-1:1,-1:1,0:edu_npat-1),  &
                   intent(in)    :: patabove
    integer(kind=int32), dimension(-2:,-2:), intent(inout) :: image

    integer :: x0, y0, npattern, n2, aver2, reste, ir, jr, kr, nevents, x1, y1
    logical :: diago, timeMode

! Identify Timing mode from patabove (patterns 4 and above are empty)

    timeMode = sum(patabove(:,:,4)) .eq. 0

! Loop over events

    nevents = size(evt)
    do kr = 1,nevents
       x0 = evt(kr)%rawx - 2          ! Center of the event
       y0 = evt(kr)%rawy - 2
       image(x0,y0) = image(x0,y0) + evt(kr)%e1
       npattern = evt(kr)%pattern
       n2 = sum(patabove(:,:,npattern))
       if (n2 > 0) then
          if ( timeMode ) then

! In Timing mode E2 is unknown. The repartition is not important

             aver2 = evt(kr)%e1 / (n2+2)
             reste = 0
             image(x0,y0) = image(x0,y0) - aver2*n2

! Patterns 2 and 3 actually bracket a series of pixels above threshold

             if (npattern == 2 .and. kr < nevents) then
                x1 = evt(kr+1)%rawx - 2          ! Center of the next event
                y1 = evt(kr+1)%rawy - 2
                if (evt(kr+1)%pattern == 3 .and. x1 > x0+3 .and. y1 == y0)  &
                     image(x0+2:x1-2,y0) = image(x0+2:x1-2,y0) + aver2
             endif
          else

! E2 is shared between pixels above threshold

             aver2 = evt(kr)%e2 / n2
             reste = evt(kr)%e2 - n2*aver2
          endif

! For large events set secondary maximum diagonal from primary
! to avoid losing the event if primary is a bad pixel

          if (n2 > 2) then
diag_loop:   do jr = -1,1,2           ! Identify diagonal pixel
                do ir = -1,1,2
                   diago = patabove(ir,jr,npattern) > 0
                   if (diago)  exit diag_loop
                enddo
             enddo diag_loop
             if (diago) then    ! Not all quadripixels have a diagonal pixel
                if (reste > 0) then
                   image(x0+ir,y0+jr) = image(x0+ir,y0+jr) + reste
                   reste = 0
                else if (evt(kr)%e1 > aver2+n2) then 
! Reduce aver2 to favor diagonal as long as it does not overtake main pixel
                   aver2 = aver2 - 1
                   image(x0+ir,y0+jr) = image(x0+ir,y0+jr) + n2
                endif
             endif
          endif
          image(x0-1:x0+1,y0-1:y0+1) = image(x0-1:x0+1,y0-1:y0+1) &
                                     + patabove(:,:,npattern) * aver2
          if (reste > 0) then
pat_loop:    do jr = -1,1
                do ir = -1,1
                   if (patabove(ir,jr,npattern) > 0) then
                      image(x0+ir,y0+jr) = image(x0+ir,y0+jr) + 1
                      reste = reste - 1
                      if (reste == 0)  exit pat_loop
                   endif
                enddo
             enddo pat_loop
          endif
       endif
    enddo

  end subroutine projectEventsEnergy

end module edusoft_mod
